#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations

__version__ = "1.0.5"
__author__ = "Decompollaborate"

from . import common
from . import elf32
from . import mips
